﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Cy4NET Programmlader Typ 1                                                          //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2017 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/boot.h>
#include <avr/pgmspace.h>
#include "AVRChip.h"
#include "TypeDefs.h"
#include "Conf.h"
#include "Board.h"
#include "Cy4Address.h"
#include "Cy4Message.h"
#include "Cy4TypePlate.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen & Definitionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#if   SPM_PAGESIZE == 32
    #define CFG_PAGE_BITS 5
#elif SPM_PAGESIZE == 64
    #define CFG_PAGE_BITS 6
#elif SPM_PAGESIZE == 128
    #define CFG_PAGE_BITS 7
#elif SPM_PAGESIZE == 256
    #define CFG_PAGE_BITS 8
#else
    #error Unsupported SPM_PAGESIZE
#endif

//////////////////////////////////////////////////////////////////////////////////////////
#define PGMLDR_START   (FLASHEND-(CFG_PGMLDR_SIZE*1024)+1)  // Programmlader Startadresse
#define PAYLOAD_CAPACITY       64 // Kapazität der PDU DAT Sektion: Fest 64 [Bytes]
#define TRANSFER_TIMEOUT      250 // Zeitfenster für Übertragung: 150 .. 250 [ms]
#define BUS_BYTE_GAP_TIMEOUT    3 // Zeitüberschreit. bei Bus-Byteübertr.: 1 ..  30 [ms]
#define DEAD_PEER_TIMEOUT    3000 // Zeitgrenze bei Ausbleiben von PDUs: 1000 .. 3000 [ms]
#define PDU_DPS              0xAA // Datagramm-Einleitungssequenz: 0b10101010
#define PDU_DIB              0x30 // Datagramm-Identifikations-Bits: 0b110000
#define PDU_DTF_P               0 // Datagram Netzlast-Flag: P (Payload)
#define PDU_DTF_D               1 // Datagram Richtungs-Flag: D (Direction)

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#if ((PGMLDR_START >> CFG_PAGE_BITS)-1) < 255   // Höchstmögliche Speicherseiten-Nummer
    typedef UInt08 PageNo;
#else
    typedef UInt16 PageNo;
#endif

#if (SPM_PAGESIZE < 0xFF)                     // Speicherindex-Typ in Abh. der Seitengröße
    typedef UInt08 PageIndex;
#else
    typedef UInt16 PageIndex;
#endif

//////////////////////////////////////////////////////////////////////////////////////////
typedef struct
{   UInt08  byte[10+PAYLOAD_CAPACITY+1];
} Cy4PDU;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt08   zero[SPM_PAGESIZE]  = { 0 };
static UInt08   page[SPM_PAGESIZE]  = { 0 };
static bool     progInProg             = false;      // true = Programmierung im Gange
static PageNo   ramPageNo              = 0;          // Nr der aktuellen Speicherseite
static          Cy4Address  selfAdr    = CY4ADDRESS_INVALID;
static Cy4PDU   pdu                    = { { 0 } };

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachricht  ///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Cy4MessageAccess
CY4PDU_Access(void)
{
    if (pdu.byte[1] == 0xC1 || pdu.byte[1] == 0xC2) return CY4MESSAGE_READ;
    else                                            return CY4MESSAGE_WRITE;
}

//////////////////////////////////////////////////////////////////////////////////////////
static Cy4Address
CY4PDU_Address(void)
{
    Cy4Address address = CY4ADDRESS(pdu.byte[2], pdu.byte[3]);
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
UInt32
CY4PDU_Location(void)
{
    return (((UInt32)pdu.byte[6])<<16) + (((UInt16)pdu.byte[7])<<8) + pdu.byte[8];
}

//////////////////////////////////////////////////////////////////////////////////////////
UInt08
CY4PDU_LocationLoByte(void)
{
    return pdu.byte[8];
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt08
CY4PDU_Count(void)
{
    return pdu.byte[9];
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt08*
CY4PDU_Data(void)
{
    return &pdu.byte[10];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Netz  ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool
CY4_IsRequest(void)
{
    return (bool)(pdu.byte[1] & BV(PDU_DTF_D));
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
CY4_HasPayload(void)
{
    return (bool)(pdu.byte[1] & BV(PDU_DTF_P));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_UpdateCRC(UInt08 data, UInt08* crc)
{
    *crc ^= data;
    for (UInt08 i=0; i<8; i++) *crc = (*crc >> 1) ^ ((*crc & 1) ? 0x8C : 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_SaveAddress(Cy4Address address)
{
    UInt08  data[] = { address.net, address.dev, ~address.net, ~address.dev };    
    BOARD_EEPSave(CFG_EEP_CY4_ADR_PARTITION, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static Cy4Address
CY4_LoadAddress(void)
{
    UInt08 data[4];
    BOARD_EEPLoad(CFG_EEP_CY4_ADR_PARTITION, data, sizeof(data));
    Cy4Address address;
    UInt08 invData[2] = { ~data[2], ~data[3] };
    if (data[0] == invData[0] && data[1] == invData[1])
    {   address.net = data[0];
        address.dev = data[1];
    }
    else
    {   address.net = CFG_DEFAULT_CY4_ADR_NET;
        address.dev = CFG_DEFAULT_CY4_ADR_DEV;
        CY4_SaveAddress(address);
    }
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_WaitForMessage(void)
{
    UInt08  i;
    UInt08  n;
    UInt08  crc;
    UInt08  byte;
    UInt16  timer = BOARD_SetTimeout(DEAD_PEER_TIMEOUT);

    do
    {   UInt08 ledPulse = 0x13;
        do
        {   if ((ledPulse&0x12) == 0x12) BOARD_InvLed();
            ledPulse++;
            BOARD_Watchdog();
            if (BOARD_Timeout(timer))
            {   // Wenn längere Zeit kein Datagramm eintrifft, Programmierung abbrechen
                progInProg = false;
            }
        } while (!BOARD_WaitForByte(&byte, 25));
        BOARD_SetLed();
        i = 0;
        n = 10 + PAYLOAD_CAPACITY;
        crc = 255;
        if (byte == PDU_DPS)
        {   do
            {   pdu.byte[i] = byte;
                if (i == 1 && (byte&0xFC) != (PDU_DIB<<2)) break;
                if (i == 9)
                {   if (byte > PAYLOAD_CAPACITY) break;
                    if (CY4_HasPayload()) n = 10+byte;
                    else                  n = 10;
                }
                CY4_UpdateCRC(byte, &crc);
                if (!BOARD_WaitForByte(&byte, BUS_BYTE_GAP_TIMEOUT)) break;
                i++;
            } while (i<n);
        }
    } while (i != n                              ||  // Anzahl empfangener Bytes korrekt?
             byte != crc                         ||  // Prüfsumme stimmt?
             !CY4_IsRequest()                    ||  // Datagramm ist eine Anfrage?
             CY4PDU_Address().net != selfAdr.net ||  // Anfrage ist ...
             CY4PDU_Address().dev != selfAdr.dev);   //            ... an uns gerichtet?
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_ReplyMessage(void)
{
    // (1) Richtungswechsel, Anfrage in Antwort verwandeln
    pdu.byte[1] ^= BV(PDU_DTF_D) + BV(PDU_DTF_P);
    // (2) Quell- und Zieladresse tauschen
    pdu.byte[2] = pdu.byte[4];
    pdu.byte[3] = pdu.byte[5];
    pdu.byte[4] = selfAdr.net;
    pdu.byte[5] = selfAdr.dev;
    // (3) Prüfsumme errechnen
    UInt08  count;
    if (CY4_HasPayload()) count = 10 + CY4PDU_Count();
    else                  count = 10;
    UInt08 crc = 0xFF;
    for (UInt08 i=0; i<count; i++) CY4_UpdateCRC(pdu.byte[i], &crc);
    pdu.byte[count] = crc;
    // (4) Sobald Bus frei ist, Datagram senden
    UInt16 timer = BOARD_SetTimeout(TRANSFER_TIMEOUT);
    UInt08 i = 0;
    do
    {   UInt08  dummy;
        if( !BOARD_WaitForByte(&dummy, BUS_BYTE_GAP_TIMEOUT) )
        {   for (i=0; i<=count; i++)
            {   UInt08 byte = pdu.byte[i];
                BOARD_TransmitByte(byte);
                UInt08  echo;
                if (!BOARD_WaitForByte(&echo, 1) || echo != byte) break;
            }
        }
    } while (i <= count && !BOARD_Timeout(timer));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ReadPage(PageNo pageNo)
{
    UInt08*  buffer;
    if (pageNo == 0) buffer = zero;
    else             buffer = page;
    UInt08* pageBaseAdr = (UInt08*)(pageNo << CFG_PAGE_BITS);
    for (PageIndex i=0; i<SPM_PAGESIZE; i++) buffer[i] = pgm_read_byte(pageBaseAdr+i);
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
PGMLDR_WritePage(PageNo pageNo)
{
    UInt08*  buffer;
    if (pageNo == 0) buffer = zero;
    else             buffer = page;
    UInt08* pageBaseAdr = (UInt08*)(pageNo << CFG_PAGE_BITS);
    // Löschen
    boot_page_erase(pageBaseAdr);
    boot_spm_busy_wait();
    // Schreiben
    for (PageIndex i=0; i<SPM_PAGESIZE; i+=2)
    {   UInt16 word = (((UInt16)buffer[i+1])<<8) + buffer[i];
        boot_page_fill(pageBaseAdr+i, word);
    }
    boot_page_write(pageBaseAdr);
    boot_spm_busy_wait();
    boot_rww_enable();
    // Prüfen
    for (PageIndex i=0; i<SPM_PAGESIZE; i++)
    {   if (buffer[i] != pgm_read_byte(pageBaseAdr+i)) return false;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_WaitForNextMessage(void)
{
    CY4_WaitForMessage();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_Setup(void)
{
    selfAdr = CY4_LoadAddress();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ProcessSystemMessage(void)
{
    if (CY4PDU_Location() < 0xFFFFF0) return;
    UInt08 locLoByte = CY4PDU_LocationLoByte(); // = (UInt08)CY4PDU_Location(); (Speicher)
    UInt08 count = CY4PDU_Count();
    if (CY4PDU_Access() == CY4MESSAGE_READ)
    {   // Call typeplate
        static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                         CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                      DEV_APPL_VERS,
                                                                      DEV_BOARD_TYPE)
                                                      };
        UInt08 i = 0;
        UInt08 k = (locLoByte-0xF0);
        do
        {   if (k >= CARDINALITY(typePlate[1])) return;
            CY4PDU_Data()[i] = pgm_read_byte(&typePlate[1][k]);
            k++;
            i++;
        } while (i<count);
        CY4_ReplyMessage();
    }
    else // CY4MSG_WRITE
    {   if (count == 1 &&
            (locLoByte == 0xF8 || locLoByte == 0xFA) &&
            CY4PDU_Data()[0] == '^')
        {   // Enter application / Enter programmloader
            if (locLoByte == 0xF8 && progInProg)
            {   PGMLDR_WritePage(ramPageNo);
                PGMLDR_WritePage(0);
            }
            CY4_ReplyMessage();
            if (locLoByte == 0xFA) BOARD_Reset();
            else                   BOARD_EnterAppl();
        }
        else if (count == 2 &&
                 locLoByte == 0xFC &&
                 (CY4PDU_Data()[0] != 0 && CY4PDU_Data()[1] != 0))
        {   // Change Cy4 address
            Cy4Address  newAdress;
            newAdress.net = CY4PDU_Data()[0];
            newAdress.dev = CY4PDU_Data()[1];
            CY4_ReplyMessage();
            selfAdr = newAdress;
            CY4_SaveAddress(newAdress);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ProcessProgrammingMessage(void)
{
    if (CY4PDU_Access() == CY4MESSAGE_WRITE && CY4PDU_Location() < PGMLDR_START)
    {   uintptr_t memAdr = (uintptr_t)CY4PDU_Location();
        if (!progInProg)
        {   ramPageNo = 0;
            PGMLDR_ReadPage(0);
            boot_page_erase_safe(0);
        }
        progInProg = true;
        for (UInt08 i=0; i<CY4PDU_Count(); i++)
        {   PageNo accPageNo = (PageNo)(memAdr >> CFG_PAGE_BITS);
            if (accPageNo != ramPageNo)
            {   if (ramPageNo > 0)
                {   if (!PGMLDR_WritePage(ramPageNo)) return;         // Schreibfehler
                }
                if (accPageNo > 0) PGMLDR_ReadPage(accPageNo);
                ramPageNo = accPageNo;
            }
            PageIndex index = (PageIndex)(memAdr & (SPM_PAGESIZE-1));
            UInt08* buffer;
            if (ramPageNo == 0) buffer = zero;
            else                buffer = page;
            buffer[index] = CY4PDU_Data()[i];
            memAdr++;
        }
        CY4_ReplyMessage();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
main(void)
{
    BOARD_Setup();
    CY4_Setup();
    while (true)
    {   PGMLDR_WaitForNextMessage();
        PGMLDR_ProcessSystemMessage();
        PGMLDR_ProcessProgrammingMessage();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
